"""A minimal set of codes, used by role provider."""


class NomenclatureCodes:
    """Codes that are used by included role providers."""

    MDC_OP_SET_ALL_ALARMS_AUDIO_PAUSE = '128284'
    MDC_OP_SET_CANCEL_ALARMS_AUDIO_PAUSE = '128285'
    MDC_OP_SET_TIME_SYNC_REF_SRC = '128505'
    MDC_ACT_SET_TIME_ZONE = '68632'
